'''
Runs server for the protein landscape computation.
'''

from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import parse_qs
from epicore_utils.modules.visualize_protein import plot_protein_landscape
import pandas as pd
import ast
import time
import numpy as np


class MyHTTPRequestHandler(BaseHTTPRequestHandler):
    ''' HTTPRequestHandler
    '''

    def do_GET(self):
        ''' GET method of the HTTPRequestHandler.
        '''        
        report = getattr(self.server, "report", "No report available")
        self.send_response(200,message='well')
        self.end_headers()
             
        if self.path == '/prot_lan.svg':
            with open('prot_lan.svg', 'rb') as f:
                self.wfile.write(f.read())
        elif 'svg' in self.path:
            with open(self.path[1:], 'rb') as f:
                self.wfile.write(f.read())
        else:
            self.wfile.write(report.encode("utf-8"))


    def do_POST(self):
        ''' POST method of the HTTPRequestHandler.
        '''
        # get accession, epicore result and proteome dictionary
        content_length = int(self.headers.get('Content-Length', 0))
        post_data = self.rfile.read(content_length) 
        parsed_data = parse_qs(post_data.decode('utf-8'))
        accession = parsed_data.get('accession', [''])[0]
        epicore_csv = parsed_data.get('epicore_csv', [''])[0]
        proteome_dict = parsed_data.get('proteome_dict', [''])[0]

        proteome_dict = ast.literal_eval(proteome_dict)

        # read in epicore result
        protein_df = pd.read_csv(epicore_csv)
    
        protein_df['grouped_peptides_start'] = protein_df['grouped_peptides_start'].apply(ast.literal_eval)
        protein_df['core_epitopes_start'] = protein_df['core_epitopes_start'].apply(lambda cell: eval(cell, {"np": np}))
        protein_df['core_epitopes_end'] = protein_df['core_epitopes_end'].apply(lambda cell: eval(cell, {"np": np}))
        protein_df['landscape'] = protein_df['landscape'].apply(ast.literal_eval)

        img_url = f'prot_lan{str(int(time.time()))}.svg'
        # compute protein landscape
        fig = plot_protein_landscape(protein_df, accession, proteome_dict)
        fig.savefig(img_url,bbox_inches='tight')
        
        self.send_response(200,message=accession)
        self.end_headers()
        self.wfile.write(img_url.encode("utf-8"))
        

    def do_OPTIONS(self):
        ''' OPTIONS method of the HTTPRequestHandler.
        '''
        self.send_response(200, "ok")
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods', 'GET,POST, OPTIONS')
        self.send_header("Access-Control-Allow-Headers", "X-Requested-With")
        self.send_header("Access-Control-Allow-Headers", "Content-Type")
        self.end_headers()


def run(port,report, server_class=HTTPServer,handler_class=MyHTTPRequestHandler):
    ''' Runs the server.

    Args:
        port: An integer defining the port.
        server_class: Specifies which server class is used.
        handler_class: Specifies which request handler class is used.
    
    '''
    server_address = ('', port)
    httpd = server_class(server_address, handler_class)
    httpd.report = report 
    httpd.serve_forever()


"""
Visualizes the landscape of a protein.
"""

import matplotlib.pyplot as plt
import matplotlib.ticker as tck
from matplotlib.ticker import MaxNLocator
import pandas as pd 
import numpy as np
import re
import ast
import logging
logger = logging.getLogger(__name__)


def plot_peptide_length_dist(first_df: pd.DataFrame, second_df: pd.DataFrame, first_explode: str, second_explode: str, first_column: str, second_column: str, first_label: str, second_label: str, mod_pattern: str) -> tuple[plt.figure,int,int]:
    """Visualize the distribution of lengths of sequences.
    
    Args:
        first_df: A pandas dataframe.
        second_df: A pandas dataframe.
        first_explode: The header of the column, that is used to explode the 
            dataframe. 
        second_column: The header of the column, that is used to explode the 
            dataframe. 
        first_column: The header of the column, that holds the sequences in 
            first_df, for which the length distribution will be plotted. 
        second_column: The header of the column, which holds the sequences in 
            second_df, for which the length distribution will be plotted.  
        first_label: Label for the values of first_column in the plot.
        second_label: Label for the values of second_column in the plot. 
        mod_pattern: A comma separated string with delimiters for peptide
            modifications


    Returns:
        A tuple including a matplotlib figure and two integers. The figure is a 
        histogram visualizing the length distribution of peptides and epitopes. 
        The two integers are the number of peptides and the number of epitopes. 
    """
    
    first_long = first_df.explode(first_explode)
    second_long = second_df.explode(second_explode)
    
    logger.info(f'{len(first_long.drop_duplicates("sequence"))} peptides were reduced to {len(second_long.drop_duplicates(["whole_epitopes", "consensus_epitopes"]))} epitopes.')

    # compute a histogram of the sequence lengths in first_column and second_column
    fig, ax = plt.subplots(layout='constrained')
    seq_first = first_long[first_column]
    seq_second = second_long[second_column]

    # remove modifications before accessing the sequence length
    pattern = r'\(.*?\)'
    peptide_seqs = seq_first.apply(lambda seq: re.sub(pattern,"",seq))
    pattern = r'\[.*?\]'
    peptide_seqs = peptide_seqs.apply(lambda seq: re.sub(pattern,"",seq))
    if mod_pattern:
        pattern = re.escape(mod_pattern.split(',')[0]) + r'.*?' + re.escape(mod_pattern.split(',')[1])
        peptide_seqs = peptide_seqs.apply(lambda seq: re.sub(pattern,"",seq))

    first_len = peptide_seqs.map(lambda pep: len(pep)).to_list()
    second_len = seq_second.map(lambda pep: len(pep)).to_list()
    
    ax.hist(first_len, bins=np.arange(5,50,1), color='grey', label=first_label, alpha=0.6)
    ax.hist(second_len, bins=np.arange(5,50,1), color='red', label=second_label, alpha=0.6)

    ax.legend()
    ax.set_xlabel('length')
    ax.set_ylabel('count')
    return fig, len(first_long), len(second_long)


def plot_protein_landscape(protein_df: pd.DataFrame, accession: str, proteome_dict: dict[str,str]) -> plt.figure:
    """Visualize the landscape of a protein.

    Args:
        protein_df: A pandas dataframe containing one protein per row.
        accession: The string of a protein accession.
        proteome_dict: A dictionary containing the reference proteome.

    Returns:
        A matplotlib bar plot that visualizes the peptide and core epitope 
        distribution across the sequence of the protein with the provided 
        accession. 
    """

    prot_row = protein_df[(protein_df['accession'] == accession)]
    if len(prot_row) == 0:
        raise Exception('The accession {} is not in your input data.'.format(accession))

    prot_seq = proteome_dict[accession]

    prot_landscape = [0 for _ in prot_seq]
    
    fig_width = max(15,round(len(prot_landscape)/50))
    fig_height = 3

    max_intens = 0

    fig, ax = plt.subplots(figsize=(fig_width, fig_height), layout='constrained')

    for group, landscape in enumerate(prot_row['landscape'].iloc[0]):

        if group % 3 == 0:
            color = 'red'
        elif group % 3 == 1:
            color = 'blue'
        else: 
            color = 'green'

        group_start = min(prot_row['grouped_peptides_start'].iloc[0][group])
        for idx, position in enumerate(landscape):
            ax.bar(group_start+int(idx),position,width=1, alpha=0.4, color=color)
        for pos in range(prot_row['core_epitopes_start'].iloc[0][group],prot_row['core_epitopes_end'].iloc[0][group]+1):
            ax.bar(pos,0.5,width=1,color=color)

        max_intens = max(max_intens, max(landscape))

    ybins=max_intens/10
    ax.yaxis.set_major_locator(MaxNLocator(nbins=max(ybins, 5)))
    ax.xaxis.set_major_locator(MaxNLocator(nbins=max(fig_width/10, 25)))

    ax.set_title('Landscape and consensus epitopes of the protein {}'.format(accession))
    ax.set_xlabel('Position in protein {}'.format(accession))
    ax.set_ylabel('Number of aligned peptides')
    return fig



def plot_core_mapping_peptides_hist(epitope_df: pd.DataFrame) ->  plt.figure:
    """A histogram of the number of peptides mapped to each epitope. 
    
    Args:
        epitope_df: A dataframe containing one epitope and 
            information of that epitope per row. 

    Returns:
        A histogram visualizing the number of peptides mapped to 
        each epitope. 
    """
    fig, ax = plt.subplots(layout='constrained')
    n_peps = epitope_df['grouped_peptides_sequence'].apply(lambda sequences: len(sequences.split(',')))
    ax.hist(n_peps,bins=np.arange(1,max(n_peps)+2,1))
    ax.set_yscale('log')
    ax.set_xlabel('number of peptides contributing to each consensus epitope')
    ax.set_ylabel('count')
    return fig


def compute_coverage(pep_start, pep_end, consensus_start, consensus_end):
    '''Compute the coverage of the consensus sequence by the peptide contributing to it.

    Args:
        peptide: Peptide sequence.
        consensus: Consensus sequence.
        entire: Sequence of entire consensus sequence.

    Returns:
        The coverage of the consensus sequence by the peptide sequence
    '''

    if pep_start < consensus_start:
        overlap = max(min(consensus_end-consensus_start+1, pep_end-consensus_start+1),0)
    else:
        overlap = max(min(pep_end-pep_start+1, consensus_end-pep_start+1),0)
    return overlap/(consensus_end-consensus_start+1)


def compute_epitopes_coverage(epitope_df: pd.DataFrame, out):
    '''Compute the coverage of consensus sequence epitopes.

    Args:
        cores_df: A pandas DataFrame containing consensus sequence epitopes..

    Returns:
        A tuple containing the list of coverages for all consensus epitope sequences and the list of coverages for consensus sequence epitope sequences that are defined by at least two peptides.
    '''
    epitope_df['grouped_peptides_start'] = epitope_df['grouped_peptides_start'].apply(lambda starts: [re.sub(r'[^a-zA-Z0-9]', '',start) for start in starts.split(',')])
    epitope_df['grouped_peptides_end'] = epitope_df['grouped_peptides_end'].apply(lambda starts: [re.sub(r'[^a-zA-Z0-9]', '',start) for start in starts.split(',')])
    epitope_df['grouped_peptides_sequence'] = epitope_df['grouped_peptides_sequence'].apply(lambda starts: [re.sub(r'[^a-zA-Z0-9]', '',start) for start in starts.split(',')])
    epitope_df = epitope_df.explode(['grouped_peptides_start', 'grouped_peptides_end', 'grouped_peptides_sequence'])
    epitope_df['coverage'] = epitope_df.apply(lambda row: compute_coverage(int(row['grouped_peptides_start']), int(row['grouped_peptides_end']), int(row['core_epitopes_start']), int(row['core_epitopes_end'])), axis=1)
    coverage_all = epitope_df['coverage'].to_list()
    epitope_df = epitope_df.drop_duplicates(['grouped_peptides_sequence', 'whole_epitopes', 'consensus_epitopes'])
    epitope_df.to_csv(f'{out}/coverage.csv')
    return coverage_all


def plot_coverage(coverage_hist_all: list[float], out:str):
    '''Plot the coverages of the consensus sequences by the peptides.

    Args:
        coverage_hist_all: List containing all coverages.
        out: Path were the plot gets saved.
    '''
    figure, axis = plt.subplots(1, 1, dpi=150)
    n_all, bins_all, _ = axis.hist(coverage_hist_all, 100, color='red', label='all consensus sequences')
    axis.set_yscale('log')
    axis.set_ylabel('Number of peptides')
    axis.set_xlabel(f'Consensus sequence coverage')
    axis.grid()
    plt.tight_layout()
    figure.savefig(f'{out}/consensus_sequence_coverage.png')


def plot_consensus_sequence_coverage(epitopes_df: pd.DataFrame, out: str):
    '''Plot the consensus sequence coverage of the computed consensus epitope sequences.
    
    Args:
        epitopes_df: A pandas dataframe containing the calculated consensus epitope sequences.
        out: Path were the plot gets saved.
    '''
    epitopes_df_copy = epitopes_df.copy()
    all = compute_epitopes_coverage(epitopes_df_copy, out)
    plot_coverage(all, out)
    logger.info(f'The calculated consensus epitope sequences have an average peptide coverage of {np.mean(all)}')


def create_html(out_name):
    ''' Create html version of plot.

    Args: 
        out_name: Location where the figure should be stored.    
    '''
    with open(f'{out_name[:-4]}svg', 'r') as svg_file:
        svg_content = svg_file.read()
        svg_content = re.sub(r'<\?xml[^>]+\?>', '', svg_content)
        svg_content = re.sub(r'<!DOCTYPE[^>]+>', '', svg_content)
        html = f'<!DOCTYPE html> <html> <body>{svg_content}</body></html>'
        with open(f'{out_name}','w') as f:
            f.write(html)


"""
Computes core epitopes, by grouping overlapping peptides, together, building a landscape for each group and identifying plateaus with a defined minimal length in each landscape. 
"""

import numpy as np
import re
import pandas as pd
from multiprocessing import get_context, cpu_count


def group_peptides_protein(row: list, min_overlap: int, max_step_size: int, intensity_column: str, total_intens: float, strict: bool) -> pd.DataFrame:
    """Group the peptides to consensus epitopes.

    Args:
        protein_df: A pandas dataframe containing one protein per row.
        min_overlap: An integer of the minimal overlap between two epitopes     
            to be grouped to the same consensus epitope.
        max_step_size: An integer of the maximal distance between the start 
            position of two epitopes to be grouped to the same consensus 
            epitope.
        intensity_column: The header of the column containing the intensities
            of the peptides.
        total_intens: The total intensity of the evidence file.
        strict: Indicates if strict version should be run.

    Returns:
        The protein_df with the five to seven additional columns: 
        grouped_peptides_start, grouped_peptides_end, 
        grouped_peptides_sequence, grouped peptides_intensity, 
        core_epitopes_intensity, relative_core_intensity and 
        sequence_group_mapping. The first five column contain the start
        position, end position, sequence and intensity of the peptides grouped
        to consensus epitopes. The column relative_core_intensity contains the 
        relative core intensities of the consensus core epitopes. The column 
        sequence group mapping contains for each peptide to which consensus 
        epitope it is grouped.
    """

    start_pos =  row['start']
    end_pos = row['end']
    sequences = row['sequence']
    samples = row['sample']
    conditions = row['condition']
    if intensity_column:
        intensity = row['intensity']
    peptide_indices = row['peptide_index']

    grouped_peptides_start = []
    grouped_peptides_end = []
    grouped_peptides_sequence = []
    grouped_peptides_sample = []
    grouped_peptides_condition = []
    grouped_peptides_max_end = 0
    grouped_peptides_index = []
    if intensity_column:
        grouped_peptides_intensity = []
        core_intensity = 0
    n_jumps = 0
    mapping = []

    for i in range(len(start_pos)-1):

        grouped_peptides_start.append(int(start_pos[i]))
        grouped_peptides_end.append(int(end_pos[i]))
        grouped_peptides_sequence.append(sequences[i])
        grouped_peptides_sample.append(samples[i])
        grouped_peptides_condition.append(conditions[i])
        grouped_peptides_max_end = max(grouped_peptides_max_end, int(end_pos[i]))
        grouped_peptides_index.append(peptide_indices[i])
        if intensity_column:
            grouped_peptides_intensity.append(intensity[i])

        step_size = int(start_pos[i+1]) - int(start_pos[i])
        mapping.append(n_jumps)
        if intensity_column:
            core_intensity += float(intensity[i])

        #first_start = grouped_peptides_start[0]
        first_end = grouped_peptides_end[0]
        group_overlap = max(min(int(end_pos[i+1])-int(start_pos[i+1])+1, first_end-int(start_pos[i+1])+1),0) 

        overlap = int(end_pos[i]) - int(start_pos[i+1]) +1
        group_overlap = min(grouped_peptides_end) - int(start_pos[i+1]) +1

        if strict:
            condition_group = (((step_size >= max_step_size) and (overlap < min_overlap)) or (overlap < min_overlap) or (group_overlap < min_overlap)) and (grouped_peptides_max_end<int(end_pos[i+1]))
        else:
            condition_group = ((step_size >= max_step_size) and (overlap < min_overlap))

        # create new peptide group after each jump
        if step_size != 0:

            if condition_group:
                row['grouped_peptides_start'].append(grouped_peptides_start)
                row['grouped_peptides_end'].append(grouped_peptides_end)
                row['grouped_peptides_sequence'].append(grouped_peptides_sequence)
                row['grouped_peptides_sample'].append(grouped_peptides_sample)
                row['grouped_peptides_condition'].append(grouped_peptides_condition)
                row['peptide_indices'].append(grouped_peptides_index)
                if intensity_column:
                    row['grouped_peptides_intensity'].append(grouped_peptides_intensity)
                    row['core_epitopes_intensity'].append(core_intensity)                   
                    
                n_jumps += 1
                grouped_peptides_end = []
                grouped_peptides_start = []
                grouped_peptides_sequence = []
                grouped_peptides_sample = []
                grouped_peptides_condition = []
                grouped_peptides_intensity = []
                grouped_peptides_index = []
                grouped_peptides_max_end = 0
                if intensity_column:
                    core_intensity = 0


    # special case for last peptide match of protein
    if len(grouped_peptides_end) == 0:

        row['grouped_peptides_start'].append([int(start_pos[-1])])
        row['grouped_peptides_end'].append([int(end_pos[-1])])
        row['grouped_peptides_sequence'].append([sequences[-1]])
        row['grouped_peptides_sample'].append([samples[-1]])
        row['grouped_peptides_condition'].append([conditions[-1]])
        row['peptide_indices'].append([peptide_indices[-1]])
        if intensity_column:
            row['grouped_peptides_intensity'].append([intensity[-1]])
            row['core_epitopes_intensity'].append(peptide_indices[-1])
        mapping.append(n_jumps)
    
    else:

        grouped_peptides_start.append(int(start_pos[-1]))
        grouped_peptides_end.append(int(end_pos[-1]))
        grouped_peptides_sequence.append(sequences[-1])
        grouped_peptides_sample.append(samples[-1])
        grouped_peptides_condition.append(conditions[-1])
        grouped_peptides_index.append(peptide_indices[-1])
        if intensity_column:
            grouped_peptides_intensity.append(intensity[-1])
            core_intensity += float(intensity[-1])
            row['grouped_peptides_intensity'].append(grouped_peptides_intensity)
            row['core_epitopes_intensity'].append(core_intensity)
            

        row['grouped_peptides_start'].append(grouped_peptides_start)
        row['grouped_peptides_end'].append(grouped_peptides_end)
        row['grouped_peptides_sequence'].append(grouped_peptides_sequence)
        row['grouped_peptides_sample'].append(grouped_peptides_sample)
        row['grouped_peptides_condition'].append(grouped_peptides_condition)
        row['peptide_indices'].append(grouped_peptides_index)
        mapping.append(n_jumps)

    row['sequence_group_mapping'] = mapping

    # update start and end position (add start and end position for multiple occurrences)
    row['start'] = [f'{start}' for group in row['grouped_peptides_start'] for start in group]
    row['end'] = [f'{end}' for group in row['grouped_peptides_end'] for end in group]
    row['peptide_index'] = [f'{index}' for group in row['peptide_indices'] for index in group]

    return row

def group_peptides(protein_df: pd.DataFrame, min_overlap: int, max_step_size: int, intensity_column: str, total_intens: float, strict: bool) -> pd.DataFrame:
    
    # start, end, sequence and intensity of peptides of one group grouped together
    protein_df['grouped_peptides_start'] = [[] for _ in range(len(protein_df))]
    protein_df['grouped_peptides_end'] = [[] for _ in range(len(protein_df))]
    protein_df['grouped_peptides_sequence'] = [[] for _ in range(len(protein_df))]
    protein_df['grouped_peptides_sample'] = [[] for _ in range(len(protein_df))]
    protein_df['grouped_peptides_condition'] = [[] for _ in range(len(protein_df))]
    if intensity_column:
        protein_df['grouped_peptides_intensity'] = [[] for _ in range(len(protein_df))]
        # for each peptide group the total and relative intensity of the entire group
        protein_df['core_epitopes_intensity'] = [[] for _ in range(len(protein_df))]
        protein_df['relative_core_intensity'] = [[] for _ in range(len(protein_df))]
    # for each peptide the index of its group
    protein_df['sequence_group_mapping'] = [[] for _ in range(len(protein_df))]
    protein_df['peptide_indices'] = [[] for _ in range(len(protein_df))]
    
    protein_df = protein_df.apply(lambda row: group_peptides_protein(row, min_overlap, max_step_size, intensity_column, total_intens, strict), axis=1)
    return protein_df


def pep_landscape(row: pd.Series) -> np.array:
    """Generate landscape of a peptide in a group.

    Args:
        row: A row of a pandas dataframe containing per row one protein, 
             peptides mapped to the protein with their start and end position 
             and their intensity. 

    Returns:
        A list containing the landscape of a protein in a given peptide group.
    """
    zeros = np.zeros([row['end_max']-row['start_min']+1])
    zeros[row['grouped_peptides_start']-row['start_min']:(row['grouped_peptides_end']-row['start_min']+1)] = 1
    return zeros

def landscape_chunk(chunk_df):
    """Generate landscapes of the peptides in chunk_df.

    Args:
        chunk_df: A subset of the protein DataFrame.

    Returns:
        The input DataFrame with the additional column pep_landscape.
    """
    chunk_df['pep_landscape'] = chunk_df.apply(lambda row: pep_landscape(row), axis=1)
    return chunk_df

def comp_landscape(protein_df: pd.DataFrame, proteome_dict: dict[str,str]) -> pd.DataFrame:
    """Compute the landscape of all consensus epitope groups.
    
    Args:
        protein_df: A pandas dataframe containing one protein per row.
        proteome_dict: A dictionary containing the reference proteome.

    Returns:
        The protein_df with two additional columns (landscape, whole_epitopes).
        The column landscape contains the landscapes of the consensus epitopes 
        group. The height of the landscape of a consensus epitope group at a 
        position is the number of peptides that contain that position. The 
        column whole_epitopes contains the entire sequence of the consensus 
        epitope group.
    """
    # get entire sequence of each group
    protein_df['start_min'] = [min(x) for x in protein_df['grouped_peptides_start']]
    protein_df['end_max'] = [max(x) for x in protein_df['grouped_peptides_end']]
    protein_df['whole_epitopes'] = protein_df['accession'].map(proteome_dict)    
    protein_df['whole_epitopes'] = [whole_epitopes[start:end+1] for start, end, whole_epitopes in zip(protein_df['start_min'],protein_df['end_max'],protein_df['whole_epitopes'])]#protein_df['whole_epitopes'].str.slice(start=protein_df['start_min'],stop=protein_df['end_max']+1)
    protein_df = protein_df.assign(whole_epitopes_all=protein_df.whole_epitopes)

    protein_df['group'] = range(len(protein_df))
    protein_df = protein_df.explode(['grouped_peptides_start', 'grouped_peptides_end', 'grouped_peptides_sample', 'grouped_peptides_condition', 'grouped_peptides_sequence'])

    n_parallel = max(1, cpu_count()-5)
    # compute the landscape for each peptide in the group
    blocksize = max(len(protein_df) // n_parallel,1)
    with get_context('spawn').Pool(min(n_parallel,blocksize)) as pool:
        chunk_dfs = pool.starmap(landscape_chunk,[(protein_df.iloc[chunk*blocksize:(chunk+1)*blocksize if chunk < min(n_parallel,blocksize) -1 else len(protein_df)],) for chunk in range(min(n_parallel,blocksize))])
    protein_df = pd.concat(chunk_dfs)

    cols = protein_df.columns.values
    aggr = {'whole_epitopes_all':lambda x: list(x), 'landscape': 'first', 'grouped_peptides_start': lambda x: list(x), 'grouped_peptides_end': lambda x: list(x), 'grouped_peptides_sample': lambda x: list(x), 'grouped_peptides_condition': lambda x: list(x), 'grouped_peptides_sequence': lambda x: list(x)}
    for col in cols:
        if col not in aggr.keys():
            aggr[col] = 'first'

    comb_landscapes = protein_df.groupby('group').agg({'pep_landscape': lambda column: np.sum(tuple(column), axis=0)}).reset_index()
    comb_landscapes = comb_landscapes.rename(columns={'pep_landscape':'landscape'})

    # compute the landscape of the entire group
    protein_df = pd.merge(protein_df, comb_landscapes, on='group')
    protein_df['landscape'] = protein_df['landscape'].apply(lambda landscape: landscape.tolist())
    protein_df = protein_df.groupby('group').agg(aggr)
    aggr['landscape'] = lambda x: list(x)
    aggr['whole_epitopes'] = lambda x: list(x)
    aggr['whole_epitopes'] = lambda x: list(x)
    del aggr['accession']
    aggr['start_min'] = lambda x: list(x)

    # combine all groups of one protein
    protein_df = protein_df.groupby('accession').agg(aggr)
    protein_df['whole_epitopes_all'] = protein_df['whole_epitopes_all'].apply(lambda x: np.hstack(x).tolist())
    protein_df = protein_df.drop(['group','pep_landscape'], axis=1)
    protein_df = protein_df.reset_index()

    return protein_df


def find_minima(landscape: np.array) -> np.array:
    """Compute local minima of a landscape.
    
    Args:
        landscape: A list containing the landscape of a peptide group.

    Returns:
        A list of the positions where the landscape has a minimum
    """
    # find minima
    landscape_f = np.roll(landscape, -1) > landscape
    landscape_b = np.roll(landscape, 1) > landscape

    # ensure minima is not at the edge
    edge = np.full(len(landscape), True)
    edge[0] = False
    edge[-1] = False
    landscape = landscape_f & landscape_b & edge
    return landscape


def new_groups(starts: list[int], ends: list[int], samples: list[str], conditions: list[str], splits: list[int], sequences: list[str]) -> list:
    """Generate new peptide groups based on the landscape minima. 

    Args:
        starts: A list containing all start positions of the current peptide group. 
        ends: A list containing all end positions of the current peptide group. 
        samples: A list containing all samples of the current peptide group. 
        conditions: A list containing all conditions of the current peptide group. 
        splits: A list containing all minima of the landscape of the current peptide group. 
        sequences: A list containing all sequences of the current peptide group.
    
    Returns:
        A list containing the new peptide groups.
    """
    start_groups = []
    end_groups = []
    sample_groups = []
    condition_groups = []
    sequence_groups = []
    start_list = []
    end_list = []
    sample_list = []
    condition_list = []
    sequence_list = []

    # if the landscape has no minima
    if len(splits) == 0:
        return [[starts], [ends], [sequences], [samples], [conditions]]

    
    for start, end, sample, condition, sequence in zip(starts, ends, samples, conditions, sequences):
        
        if len(splits) == 0 or splits[0] > start:
            start_list.append(start)
            end_list.append(end)
            sample_list.append(sample)
            condition_list.append(condition)
            sequence_list.append(sequence)

        elif splits[0] <= start: # TODO check if split position is correct here (right index and < or <=)
            # start new peptide group after minimum
            start_groups.append(start_list)
            end_groups.append(end_list)
            sample_groups.append(sample_list)
            condition_groups.append(condition_list)
            sequence_groups.append(sequence_list)
            sequence_list = [sequence]
            start_list = [start]
            end_list = [end]
            sample_list = [sample]
            condition_list = [condition]
            splits = splits[1:]

    start_groups.append(start_list)
    end_groups.append(end_list)
    sample_groups.append(sample_list)
    condition_groups.append(condition_list)
    sequence_groups.append(sequence_list)

    return [start_groups, end_groups, sequence_groups, sample_groups, condition_groups]


def group_refinement(protein_df: pd.DataFrame, proteome_dict: dict[str,str]):
    """Split peptide groups at landscape minima.

    Args: 
        protein_df: A pandas dataframe containing one protein per row.
        proteome_dict: A dictionary containing the reference proteome.

    Returns:
        The pandas dataframe with refined peptide groups.
    """

    # convert protein_df so each peptide group is in one row
    protein_df = protein_df.explode(['grouped_peptides_start', 'grouped_peptides_end', 'grouped_peptides_sample', 'grouped_peptides_condition', 'grouped_peptides_sequence', 'landscape', 'start_min'])
    
    # find landscape minima
    protein_df['split_rows'] = protein_df['landscape'].apply(lambda landscape: find_minima(landscape))
    protein_df['split_position'] = protein_df.apply(lambda row: np.where(row['split_rows'])[0]+row['start_min'], axis=1)

    # refine peptide groups
    protein_df[['new_groups_start', 'new_groups_end', 'new_group_sequences', 'new_groups_sample', 'new_groups_condition']] = protein_df.apply(lambda row: new_groups(row['grouped_peptides_start'], row['grouped_peptides_end'], row['grouped_peptides_sample'], row['grouped_peptides_condition'], row['split_position'], row['grouped_peptides_sequence']), axis=1, result_type='expand')
    protein_df = protein_df.explode(['new_groups_start', 'new_groups_end', 'new_groups_sample', 'new_groups_condition', 'new_group_sequences'])
    protein_df = protein_df.drop(['grouped_peptides_start', 'grouped_peptides_end', 'grouped_peptides_sample', 'grouped_peptides_condition', 'grouped_peptides_sequence'], axis=1)
    protein_df = protein_df.rename(columns={'new_groups_start':'grouped_peptides_start', 'new_groups_end':'grouped_peptides_end', 'new_groups_sample':'grouped_peptides_sample','new_groups_condition':'grouped_peptides_condition', 'new_group_sequences': 'grouped_peptides_sequence'})
    protein_df = protein_df.drop(['landscape'], axis=1)

    # recompute group landscapes
    #protein_df = protein_df.group_by(['accession'])
    protein_df = comp_landscape(protein_df,proteome_dict)

    return protein_df


def get_consensus_epitopes(protein_df: pd.DataFrame, min_epi_len: int) -> pd.DataFrame:
    """Compute the consensus epitope sequence of each consensus epitope group.
    
    Args:
        protein_df: A pandas dataframe containing one protein per row.
        min_epi_length: An integer of the minimal length of a consensus epitope.

    Returns:
        The protein_df with one additional column, that contains the consensus epitope sequence of each consensus epitope group.
    """
    protein_df['consensus_epitopes'] = [[] for _ in range(len(protein_df))]
    protein_df['consensus_epitopes_all'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_start'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_end'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_start_all'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_end_all'] = [[] for _ in range(len(protein_df))]

    for r, row in protein_df.iterrows():
        for group,landscape in enumerate(row['landscape']):
            
            # build consensus epitopes
            total_counts = np.unique(landscape)
            total_counts[::-1].sort()
        
            # find total coverage for which consensus epitope is at least min_epi_len long
            for total_count in total_counts:

                Z = landscape < total_count

                # get lengths of peptide sequences with coverage above the current threshold
                seqs_idx = np.where(np.diff(np.hstack(([False],~Z,[False]))))[0].reshape(-1,2)
                
                # get length of longest peptide subsequences with current count
                ce_start_pos = seqs_idx[np.diff(seqs_idx, axis=1).argmax(),0]
                current_pep_length = np.diff(seqs_idx, axis=1).max()
                
                # check if min_epi_length is fulfilled for that sequence
                if current_pep_length >= min_epi_len:

                    # get position of epitope in protein sequences
                    pep_in_prot_start = ce_start_pos.item()
                    pep_in_prot_end = pep_in_prot_start + current_pep_length.item()

                    # get consensus epitopes
                    whole_epitope_wo_mod = protein_df.at[r,'whole_epitopes'][group]
                    for _ in row['grouped_peptides_sequence'][group]:
                        protein_df.at[r,'consensus_epitopes_all'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                        protein_df.at[r,'core_epitopes_start_all'].append((pep_in_prot_start+min(row['grouped_peptides_start'][group])))
                        protein_df.at[r,'core_epitopes_end_all'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1) 
                    protein_df.at[r,'consensus_epitopes'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                    protein_df.at[r,'core_epitopes_start'].append(pep_in_prot_start+min(row['grouped_peptides_start'][group]))
                    protein_df.at[r,'core_epitopes_end'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1)
                    break
                
                # if no core with length > min_epi_length
                if total_count == total_counts[-1]:
                    pep_in_prot_start = ce_start_pos.item()
                    pep_in_prot_end = pep_in_prot_start + current_pep_length.item()
                    whole_epitope_wo_mod = protein_df.at[r,'whole_epitopes'][group]
                    for _ in row['grouped_peptides_sequence'][group]:
                        protein_df.at[r,'consensus_epitopes_all'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                        protein_df.at[r,'core_epitopes_start_all'].append(pep_in_prot_start+min(row['grouped_peptides_start'][group]))
                        protein_df.at[r,'core_epitopes_end_all'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1)
                    protein_df.at[r,'consensus_epitopes'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                    protein_df.at[r,'core_epitopes_start'].append(pep_in_prot_start+min(row['grouped_peptides_start'][group]))
                    protein_df.at[r,'core_epitopes_end'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1)
                    
    protein_df['proteome_occurrence'] = protein_df.apply(lambda row: [row['accession']+':'+str(row['consensus_epitopes_all'][i])+':'+str(row['core_epitopes_start_all'][i])+'-'+str(row['core_epitopes_end_all'][i]) for i in range(len(row['core_epitopes_start_all']))],axis=1)
    return protein_df


def reorder_peptides(row: pd.Series, intensity_column: str) -> pd.Series:
    """Reorder the peptides mapped to a protein by their position.
    
    Args:
        row: A row of a pandas dataframe containing per row one protein, 
            peptides mapped to the protein, start and end position of the peptide in the protein and the intensity of the peptide.
        intensity_column: The header of the column containing the intensities
            of the peptides.
    Returns:
        A reordered version of the input row, where the start positions are sorted in ascending order, the indices of the other columns are reordered in the same pattern. 
    """
    if intensity_column:
        lists = list(zip(row['start'], row['end'], row['sequence'], row['intensity'], row['peptide_index'], row['sample'], row['condition']))
        lists = sorted(lists, key=lambda x: x[5], reverse=True)
        lists = sorted(lists, key=lambda x: int(x[1]), reverse=True)
        lists = sorted(lists, key=lambda x: x[2], reverse=True)
        sorted_lists = sorted(lists, key=lambda x: int(x[0]))
        starts, ends, sequences, intensities, indices, samples, conditions = zip(*sorted_lists)
        return list(starts), list(ends), list(sequences), list(intensities), list(indices), list(samples), list(conditions)
    else:
        lists = list(zip(row['start'], row['end'], row['sequence'], row['peptide_index'], row['sample'], row['condition']))
        lists = sorted(lists, key=lambda x: x[4], reverse=True)
        lists = sorted(lists, key=lambda x: x[2], reverse=True)
        lists = sorted(lists, key=lambda x: int(x[1]), reverse=True)
        sorted_lists = sorted(lists, key=lambda x: int(x[0]))
        starts, ends, sequences, indices, samples, conditions = zip(*sorted_lists)
        return list(starts), list(ends), list(sequences), list(indices), list(samples), list(conditions)


def compute_consensus_epitopes(protein_df: pd.DataFrame, min_overlap: int, max_step_size: int, min_epi_len: int, intensity_column: float, mod_pattern: str, proteome_dict: dict[str,str], total_intens: float, strict: bool) -> pd.DataFrame:
    """ Compute the core and whole sequence of all consensus epitope groups. 
    
    Args:
        protein_df: A pandas dataframe containing one protein per row.
        min_overlap: An integer of the minimal overlap between two epitopes     
            to be grouped to the same consensus epitope.
        max_step_size: An integer of the maximal distance between the start 
            position of two epitopes to be grouped to the same consensus 
            epitope.
        min_epi_len: An integer of the minimal length of a consensus epitope.
        intensity_column: The header of the column containing the intensities
            of the peptides.
        mod_pattern: A comma separated string with delimiters for peptide
            modifications
        proteome_dict: A dictionary containing the reference proteome.
        total_intens: The total intensity of the evidence file.
        strict: A boolean indicating if the strict version should be run.

    Returns:
        The protein_df containing for each protein the core and whole sequence of each of its consensus epitope groups.
    """
    if intensity_column:
        protein_df[['start', 'end', 'sequence', 'intensity', 'peptide_index', 'sample', 'condition']] = protein_df.apply(lambda row: pd.Series(reorder_peptides(row, intensity_column)), axis=1)
    else:
        protein_df[['start', 'end', 'sequence', 'peptide_index', 'sample', 'condition']] = protein_df.apply(lambda row: pd.Series(reorder_peptides(row, intensity_column)), axis=1)
    # group peptides
    protein_df = group_peptides(protein_df, min_overlap, max_step_size, intensity_column, total_intens, strict)
    protein_df = protein_df.explode(['grouped_peptides_start', 'grouped_peptides_end', 'grouped_peptides_sample', 'grouped_peptides_sequence', 'grouped_peptides_condition'])
    protein_df = comp_landscape(protein_df, proteome_dict)
    if not strict:
        protein_df = protein_df[['accession','sequence','start','end','peptide_index','sample','condition','grouped_peptides_start','grouped_peptides_end','grouped_peptides_sequence','grouped_peptides_sample', 'grouped_peptides_condition','sequence_group_mapping','landscape','whole_epitopes','whole_epitopes_all', 'start_min']]
        protein_df = group_refinement(protein_df, proteome_dict)
    protein_df = protein_df[['accession','sequence','start','end','peptide_index','sample','condition','grouped_peptides_start','grouped_peptides_end','grouped_peptides_sequence','grouped_peptides_sample', 'grouped_peptides_condition','sequence_group_mapping','landscape','whole_epitopes','whole_epitopes_all']]
    protein_df = get_consensus_epitopes(protein_df, min_epi_len)
    return protein_df



"""
Reads in the evidence file and reference profile, computes the peptides positions in the proteome and links a protein accession with all peptides associated with the protein. 
"""

import pandas as pd 
import re
import numpy as np 
from Bio import SeqIO
import os 
import itertools
import polars as pl
from multiprocessing import get_context, cpu_count
from typing import Union


import logging
logger = logging.getLogger(__name__)


def read_entire_id_output(id_output: str, polars: bool) -> Union[pd.DataFrame, pl.DataFrame]:
    '''Read in entire evidence file.

    Args:
        id_output: Path to the evidence file.
        polars: Indicates if csv should be loaded in polars or pandas.

    Returns:
        The evidence file in a polars or pandas dataframe.
    '''
    if polars:
        # determine the file type
        ext = os.path.splitext(id_output)[1]
        if ext == '.csv':
            peptides_df = pl.read_csv(id_output, separator=',', infer_schema_length=0).with_row_index('peptide_index')
        elif ext == '.tsv':
            peptides_df = pl.read_csv(id_output, separator='\t', infer_schema_length=0).with_row_index('peptide_index')
        elif ext == '.xlsx':
            peptides_df = pl.read_excel(id_output)
        else:
            raise Exception('The file type of your evidence file is not supported. \
                            Please use an evidence file that has one of the \
                            following file types: csv, tsv, xlsx')
        
    else:
        ext = os.path.splitext(id_output)[1]
        if ext == '.csv':
            peptides_df = pd.read_csv(id_output, delimiter=',')
        elif ext == '.tsv':
            peptides_df = pd.read_csv(id_output, delimiter='\t')
        elif ext == '.xlsx':
            peptides_df = pd.read_excel(id_output)
        else:
            raise Exception('The file type of your evidence file is not supported. \
                            Please use an evidence file that has one of the following \
                            file types: csv, tsv, xlsx')
    
    return peptides_df



def read_id_output(id_output: str, seq_column: str, protacc_column: str, intensity_column: str, 
                   start_column: str, end_column: str, delimiter: str, sample_column: str, 
                   condition_column: str, polars: bool) -> Union[pd.DataFrame, pl.DataFrame]:
    """Read in the evidence file.

    Args:
        id_output: The string of the path to the evidence file.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        delimiter: The delimiter that separates multiple entries in one column 
            in the evidence file.
        sample_column: The header of the column containing the sample 
            information.
        condition_column: The header of the column containing the condition 
            information.

    Returns:
        A polars dataframe containing the columns sequence, protein accession, 
        peptides intensity, sample, condition and optional the columns start 
        and end of the input evidence file.

    Raises:
        Exception: If the file type of the provided evidence file is not 
            supported.
        Exception: If the provided column does not exist in the provided 
            evidence file.
        Exception: If a mandatory column header is not provided.
    """
    peptides_df = read_entire_id_output(id_output, True)
    # check that the mandatory headers are provided and all provided column 
    # headers are part of the evidence file
    if seq_column not in peptides_df.columns:
        if seq_column:
            raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(seq_column))
        else:
            raise Exception('The header for the column containing the peptide \
                            sequences in the evidence file is mandatory. \
                            Please provide the correct header name.')
    if protacc_column not in peptides_df.columns:
        if protacc_column:
            raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(protacc_column))
        else:
            raise Exception('The header for the column containing the protein \
                            accessions in the evidence file is mandatory. \
                            Please provide the correct header name.')
    if (intensity_column not in peptides_df.columns) and intensity_column:
        raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(intensity_column))
    if (start_column not in peptides_df.columns) and start_column:
        raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(start_column))
    if (end_column not in peptides_df.columns) and end_column:
        raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(end_column))

    # read in evidence file (start, stop position and intensity only when 
    # provided)
    if start_column and end_column:
        if intensity_column:
            peptides_df = peptides_df.select(protacc_column, seq_column, intensity_column, 'peptide_index', start_column, end_column, sample_column, condition_column)
        else:
            peptides_df = peptides_df.select(protacc_column, seq_column, 'peptide_index', start_column, end_column, sample_column, condition_column)

        # split if peptide occurs multiple times in proteome
        peptides_df = peptides_df.with_columns(pl.col(start_column).cast(pl.String))
        peptides_df = peptides_df.with_columns(pl.col(end_column).cast(pl.String))
        peptides_df = peptides_df.with_columns((pl.col(start_column).str.split(delimiter).alias(start_column)))
        peptides_df = peptides_df.with_columns((pl.col(end_column).str.split(delimiter).alias(end_column)))
        peptides_df = peptides_df.with_columns((pl.col(protacc_column).str.split(delimiter).alias(protacc_column)))
    
    else:
        if intensity_column:
            peptides_df = peptides_df.select(protacc_column, seq_column, intensity_column, 'peptide_index', sample_column, condition_column)
        else:
            peptides_df = peptides_df.select(protacc_column, seq_column, 'peptide_index', sample_column, condition_column)
        
        # split accessions if peptide occurs multiple times in proteome
        peptides_df = peptides_df.with_columns((pl.col(protacc_column).str.split(delimiter)).alias(protacc_column))

    return peptides_df



def proteome_to_dict(proteome: str) -> dict[str,str]:
    """Read reference proteome into dictionary.

    Args:
        proteome: The string of the path to the reference proteome.

    Returns: 
        The reference proteome as a dictionary.
    """
    proteome_dict = {}
    proteome = SeqIO.parse(open(proteome),'fasta')
    for protein in proteome:
        #proteome_dict[protein.id.split('|')[1]] = str(protein.seq)
        proteome_dict[protein.id] = str(protein.seq)
    return proteome_dict


def get_pos(accessions_s: list[str], peptides: str, fasta_dict: dict) -> list[list]:
    '''Compute positions of peptide in sequence.

    Args:
        accessions: List of protein accessions to which peptide is mapped.
        peptide: Peptide sequence of interest.
        fasta_dict: Dictionary containing reference proteome.

    Returns:
        Lists containing the accessions and start and end position of a peptide in the proteome.
    '''

    positions = []
    for accessions, peptide in zip(accessions_s,peptides):
        starts = ''
        ends = ''
        accessions_str = ''
        for accession in accessions:
            if accession != 'unmapped':
                seq = fasta_dict[accession]
                groups_pos = re.finditer(f'(?=({peptide}))', seq)
                for group_pos in groups_pos:
                    pep_start, pep_end = group_pos.span()
                    accessions_str += f';{accession}'
                    starts += f';{pep_start}'
                    ends += f';{pep_end+len(peptide)-1}'
            else:
                accessions_str = f';unmapped'

        
        positions.append(f'{accessions_str[1:]}~{starts[1:]}~{ends[1:]}')
    return positions


def add_positions(fasta_dict: dict, peptides_df: pl.DataFrame) -> pl.DataFrame:
    '''Adds positions of peptides in the proteins.

    Args:
        fasta_dict: Dictionary containing peptides as keys and accessions as values.
        peptides_df: Polars dataframe containing the peptide sequence and the proteins they map to.

    Returns:
        The input dataframe with peptide positions.
    '''

    peptides_df = peptides_df.with_columns(pl.struct('accessions','sequence').map_batches(lambda x: pl.Series(get_pos(x.struct.field('accessions'), x.struct.field('sequence'), fasta_dict)),return_dtype=pl.String).str.split('~').alias('positions'))
    peptides_df = peptides_df.with_columns(pl.col('positions').list.get(0).alias('accessions'))
    peptides_df = peptides_df.with_columns(pl.col('positions').list.get(1).alias('start'))
    peptides_df = peptides_df.with_columns(pl.col('positions').list.get(2).alias('end'))
    peptides_df = peptides_df.with_columns(pl.col('accessions').str.split(';').alias('accessions'))
    peptides_df = peptides_df.with_columns(pl.col('start').str.split(';').alias('start'))
    peptides_df = peptides_df.with_columns(pl.col('end').str.split(';').alias('end'))

    return peptides_df



def group_repetitive(start: list[int], end: list[int], pep: str, acc: str, idx: list[int], sample: list[str], condition: list[str])-> str:
    """Group peptide occurrences that belong to the same repetitive region.

    Args:
        start: List of all start positions of the peptides in the protein.
        end: List of all end positions of the peptide in the protein.
        pep: The peptide sequence.
        acc: The protein accession.
        idx: List of all peptide indices.
        sample: List of all samples of the peptides in the protein.
        condition: List of all conditions of the peptides in the protein.

    Returns:
        Returns a tuple (starts,ends), where starts is a list containing the
        updated start positions and ends is a list of updated end positions. 
        These contain from the input start and end positions all start and end 
        positions that are not part of a repetitive region. For the start and 
        end positions that are part of repetitive regions the lowest start 
        position and highest end position is kept for each repetitive region. 
    """
    #lodo have to be sorted


    current = -1
    updated_start = []
    updated_end = []
    updated_idx = []
    updated_conditions = []
    updated_peps = []
    updated_samples = []

    lists = list(zip(start, end, idx))
    lists = sorted(lists, key=lambda x: int(x[0]))
    start, end, idx = zip(*lists)
    group_ends = []
    # add the first occurrences start positions to the start positions
    for i in idx[0]:
        updated_start.append(str(start[0]))

    for pep_pos in range(len(start)-1):
        for i in idx[pep_pos]:
            group_ends.append(end[pep_pos])
        # two start positions are not part of one repetitive region if the next start position is higher than the current end position 
        if int(start[pep_pos + 1]) > int(end[pep_pos]): # new group
            for i in idx[pep_pos]:                    
                updated_idx.append(str(i))
                updated_peps.append(pep)
                updated_samples.append(sample[0])
                updated_conditions.append(condition[0])
                current = pep_pos
            for i in idx[pep_pos+1]:
                updated_start.append(str(start[pep_pos+1]))
            # add max end of repetitive group
            for _ in group_ends:
                    updated_end.append(str(max(group_ends)))
            group_ends = []

        else: #  add min_start for repetitive group
            for i in idx[pep_pos]:
                updated_idx.append(str(i))
                updated_peps.append(pep)
                updated_samples.append(sample[0])
                updated_conditions.append(condition[0])
            for i in idx[pep_pos+1]:
                updated_start.append(str(start[current+1]))
    for i in idx[-1]:
        group_ends.append(end[-1])
    # add the last occurrences end position to the end positions
    for i in idx[-1]:
        updated_idx.append(str(i))
        updated_peps.append(pep)
        updated_samples.append(sample[0])
        updated_conditions.append(condition[0])
    for _ in group_ends:
        updated_end.append(str(max(group_ends)))

    # reduce each occurrence to one
    updated_df = pd.DataFrame({'start':updated_start, 'end':updated_end, 'peps':updated_peps, 'idx':updated_idx, 'sample':updated_samples, 'cond':updated_conditions})
    updated_df = updated_df.drop_duplicates()
    updated_start = ';'.join(updated_df['start'])
    updated_end = ';'.join(updated_df['end'])
    updated_peps = ';'.join(updated_df['peps'])
    updated_idx = ';'.join(updated_df['idx'])
    updated_samples = ';'.join(updated_df['sample'])
    updated_conditions = ';'.join(updated_df['cond'])

    return f'{updated_start}|{updated_end}|{updated_idx}|{updated_peps}|{updated_samples}|{updated_conditions}'

            
def group_repetitive_chunk(chunk_df: pl.DataFrame, start: int, end: int):
    ''' Group repetitive peptides.

    chunk_df: A polars DataFrame containing one protein per row.
    start: The row, at which the polars DataFrame gets sliced.
    end: The row, at which the polars DataFrame gets sliced.

    Returns:
        A slice of the input DataFrame with grouped repetitive peptides.
    '''
    return chunk_df[start:end].with_columns(pl.struct('start', 'end', 'sequence', 'accessions', 'peptide_index', 'sample', 'condition').map_elements(lambda x: group_repetitive(x['start'], x['end'], x['sequence'], x['accessions'], x['peptide_index'], x['sample'], x['condition']), return_dtype=pl.String).str.split('|').alias('repetitive'))


def prot_pep_link(peptides_df: pd.DataFrame, seq_column: str, protacc_column: str, intensity_column: str, start_column: str, end_column: str, delimiter: str, sample_column: str, condition_column: str) -> pl.DataFrame:
    """Converts a dataframe from one peptide per row to one protein per row.
    
    Args:
        peptides_df: A pandas dataframe containing the columns peptide sequence,
            protein accession, intensity, start position and end position.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        delimiter: The delimiter that separates multiple entries in one column 
            in the evidence file.
        sample_column: The header of the column containing sample information.
        condition_column: The header of the column containing condition 
            information.
    
    Returns:
        A polars dataframe containing one protein per row and all peptides 
        mapped to that protein in the peptides_df, with their start position, 
        end position and intensity.

    Raises:
        Exception: If a peptide does not occur in the protein to which it is 
            mapped.
    """
    start_column = 'start'
    end_column='end'

    if intensity_column:
        proteins = pd.DataFrame(columns=['accession', 'sequence', 'intensity', 'start','end', 'peptide_index'])
    else:
        proteins_df = peptides_df.explode(protacc_column, start_column, end_column)
        proteins_df = proteins_df.with_columns(pl.col(start_column).cast(pl.Int64))
        proteins_df = proteins_df.sort(start_column).group_by(protacc_column).agg(pl.col(seq_column), pl.col(start_column), pl.col(end_column), pl.col('peptide_index'), pl.col(sample_column), pl.col(condition_column))
        proteins_df = proteins_df.explode(start_column, end_column, seq_column, 'peptide_index', sample_column, condition_column).group_by(seq_column, protacc_column, sample_column, condition_column).agg(pl.col(start_column), pl.col(end_column), pl.col('peptide_index'))
        proteins_df = proteins_df.with_columns(pl.col(end_column).cast(pl.List(pl.Int64)))
        proteins_df = proteins_df.with_columns(pl.col(start_column).cast(pl.List(pl.Int64)))
        proteins_df = proteins_df.with_columns(pl.col('peptide_index').cast(pl.List(pl.List(pl.Int64))))
        proteins_df = proteins_df.with_columns(pl.col('sample').cast(pl.List(pl.String)))
        proteins_df = proteins_df.with_columns(pl.col('condition').cast(pl.List(pl.String)))

        n_parallel = max(1, cpu_count()-5)
        block_size = max(len(proteins_df) // n_parallel,1)
        with get_context('spawn').Pool(n_parallel) as pool:
            chunk_dfs = pool.starmap(group_repetitive_chunk,[(proteins_df,chunk*block_size,(chunk+1)*block_size if chunk < (n_parallel-1) else len(proteins_df)) for chunk in range(n_parallel)])
        proteins_df = pl.concat(chunk_dfs)
        proteins_df = proteins_df.with_columns(pl.col('repetitive').list.get(0).alias('start'))
        proteins_df = proteins_df.with_columns(pl.col('repetitive').list.get(1).alias('end'))
        proteins_df = proteins_df.with_columns(pl.col('repetitive').list.get(2).alias('peptide_index'))
        proteins_df = proteins_df.with_columns(pl.col('repetitive').list.get(3).alias('sequence'))
        proteins_df = proteins_df.with_columns(pl.col('repetitive').list.get(4).alias('sample'))
        proteins_df = proteins_df.with_columns(pl.col('repetitive').list.get(5).alias('condition'))
        proteins_df = proteins_df.with_columns((pl.col(start_column).str.split(delimiter)).alias(start_column))
        proteins_df = proteins_df.with_columns((pl.col(end_column).str.split(delimiter)).alias(end_column))
        proteins_df = proteins_df.with_columns((pl.col('peptide_index').str.split(delimiter)).alias('peptide_index'))
        proteins_df = proteins_df.with_columns((pl.col('sequence').str.split(delimiter)).alias('sequence'))
        proteins_df = proteins_df.with_columns((pl.col('sample').str.split(delimiter)).alias('sample'))
        proteins_df = proteins_df.with_columns((pl.col('condition').str.split(delimiter)).alias('condition'))
        proteins_df = proteins_df.explode('start', 'end', 'peptide_index','sequence', 'sample', 'condition')
        proteins_df = proteins_df.group_by(protacc_column).agg(pl.col(seq_column), pl.col('start'), pl.col('end'), pl.col('peptide_index'), pl.col('sample'), pl.col('condition'))
        proteins_df = proteins_df.rename({protacc_column:'accession'})

    return proteins_df


def parse_input(evidence_file: str, seq_column: str, protacc_column: str, intensity_column: str, start_column: str, end_column: str, delimiter: str, proteome_dict: dict[str,str], mod_pattern: str, sample_column: str, condition_column: str) -> tuple[pd.DataFrame, int, float]:
    """Parse the evidence file.
    
    Args:
        evidence_file: The string of the path to the evidence file.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        delimiter: The delimiter that separates multiple entries in one column 
            in the evidence file.
        proteome_dict: A dictionary containing the reference proteome.
        mod_pattern: A comma separated string with delimiters for peptide
            modifications
    
    Returns:
        A tuple of a pandas dataframe, an integer and a float. The dataframe contains 
        one protein per row and all peptides mapped to that protein in the 
        peptides_df, with their start position, end position and intensity.
        The integer is the number of peptides that are removed due to their 
        accession not appearing in the proteome file. The float is the total intensity 
        of the evidence file. 
    """
    peptides_df = read_id_output(evidence_file, seq_column, protacc_column, intensity_column, start_column, end_column, delimiter, sample_column, condition_column, True)
    # get peptides/proteins with protein accessions that do not appear in the proteome
    peptides = pl.Series(peptides_df.with_columns((pl.col(protacc_column).list.filter(~pl.element().is_in(list(proteome_dict.keys())))).alias('removed')).select('removed')).to_list()
    n_removed_proteins = set(itertools.chain.from_iterable(peptides))
    #n_removed_proteins = set(itertools.chain.from_iterable(pep for pep in peptides if pep))

    # remove all peptides occurring multiple times in different modification and charge states
    peptides_df = peptides_df.with_columns((pl.col(seq_column).str.replace_all(r'\(.*?\)', '')).alias(seq_column))
    # remove peptides with protein accessions that do not appear in the proteome 
    if start_column and end_column:
        peptides_df = peptides_df.with_columns(pl.col(start_column).list.gather(pl.col(protacc_column).list.eval(pl.arg_where(~pl.element().is_in(n_removed_proteins)).alias(start_column))))
        peptides_df = peptides_df.with_columns(pl.col(end_column).list.gather(pl.col(protacc_column).list.eval(pl.arg_where(~pl.element().is_in(n_removed_proteins)).alias(end_column))))
        peptides_df = peptides_df.group_by([protacc_column, seq_column, sample_column, condition_column]).agg(pl.col(start_column).first(), pl.col(end_column).first(), pl.col('peptide_index'))
    else:
        peptides_df = peptides_df.group_by([protacc_column, seq_column, sample_column, condition_column]).agg(pl.col('peptide_index'))
    peptides_df = peptides_df.with_columns(pl.col(protacc_column).list.gather(pl.col(protacc_column).list.eval(pl.arg_where(~pl.element().is_in(n_removed_proteins)).alias(protacc_column))))
    # remove peptides that are not annotated with any proteome accession
    peptides_df = peptides_df.remove(pl.col(protacc_column).list.len() == 0)

    # compute start and end positions of the peptides
    if not start_column and not end_column:
        peptides_df = add_positions(proteome_dict, peptides_df)

    logger.info(f'Peptides mapped to the following {len(n_removed_proteins)} proteins were removed since the proteins do not appear in the proteome fasta file: {n_removed_proteins}.')
    protein_df = prot_pep_link(peptides_df, seq_column, protacc_column, intensity_column, start_column, end_column, delimiter, sample_column, condition_column)
    if intensity_column:
        total_intens = peptides_df[intensity_column].sum()
    else:
        total_intens = 0

    return protein_df, len(n_removed_proteins), total_intens


'''
Generates a report for the results of epicore.
'''

import webbrowser 
import ast
import pandas as pd
import threading
import epicore_utils.modules.locserver


def gen_report(length_distribution: str, intensity_hist: str, epitope_df: pd.DataFrame, peps: int, epitopes: int, n_removed_peps: int, ctx, evidence_file: str, epicore_csv: str):
    """ Generates a report including the most important information of the results. 

    Args:
        length_distribution: The path to the file containing the length 
            distribution of peptides and epitopes. 
        intensity_hist: The path to the file containing a histogram of the 
            number of peptides contributing to each core. 
        epitope_df: A dataframe containing one epitope and information of that  
            epitope per row. 
        peps: Number of peptides in the evidence file. 
        epitopes: Number of epitopes generated.
        n_removed_peps: Number of peptides removed from the evidence file due 
            to the absence of their accessions in the proteome.
        ctx: Object containing the input parameters.
        evidence_file: Path to the evidence file.
        epicore_csv: Path to the epicore result csv.

    Returns:
        Open a html report summarizing some information of the epicore 
        result. For example the report includes a figure of the peptide/epitope 
        length distribution, a histogram of the number of peptides mapped to 
        each epitope and the ten epitopes with the highest number of mapped 
        peptides. 
    """

    # compute mean of epitope core sequence
    len_core_epitopes = epitope_df['consensus_epitopes'].apply(lambda sequence: len(sequence))
    mean_core_length = sum(len_core_epitopes) / len(len_core_epitopes)

    # compute mean of epitope whole sequence
    len_whole_epitopes = epitope_df['whole_epitopes'].apply(lambda sequence: len(sequence))
    mean_whole_length = sum(len_whole_epitopes) / len(len_whole_epitopes)

    epitope_df_sort = epitope_df.sort_values(by='landscape', key=lambda landscapes: landscapes.apply(lambda landscape: max(ast.literal_eval(landscape))), ascending=False).head(10)
    epitope_df_sort_html = epitope_df_sort.to_html()

    # get accession of top ten epitopes to add them as options to the combobox
    epitope_df_sort['accession'] = epitope_df_sort['accession'].apply(lambda row: row.split(','))
    epitope_df_sort_accessions = epitope_df_sort['accession'].explode().tolist() 

    html_content = f''' <!DOCTYPE html>
                        <html>
                        <head>
                        <script src="https://cdn.jsdelivr.net/npm/brython@3/brython.min.js">
                        </script>
                        <script src="https://cdn.jsdelivr.net/npm/brython@3/brython_stdlib.js">
                        </script>
                            <style> 
                                .row {{
                                    display: flex;
                                    justify-content: space-between;
                                }}
                                .column {{
                                    flex: 1;
                                    margin: 0 10px;
                                }}
                                h1{{text-align: center;}}
                            </style>
                            <title>Epicore report</title>
                        </head>
                        <body>
                            <h1>Epicore report</h1>
                            <div class="row">
                                <div class="column"><p> The histogram shows the number of peptides/epitopes of a certain length. {peps} peptides were reduced to {epitopes} epitopes.</p><img src="{length_distribution}"></div>
                                <div class="column"><p> The histogram visualizes how many peptides contribute to the different epitopes.<br></p><img src="{intensity_hist}"></div>
                            </div>

                            <script type="text/python">
                                from browser import document, bind, ajax, html

                                select_element = document["exampleList"]
                                for item in {epitope_df_sort_accessions}:
                                    option = html.OPTION(value=item)
                                    select_element <= option

                                def on_complete(req):
                                    if req.status == 200 or req.status == 0:
                                        image_url = req.text  
                                        img_element = document["image_container"]
                                        img_element.clear()
                                        img_element <= html.IMG(src=image_url, alt="The accession you selected is not in your input data. Try a different accession.", style="width: 1400px;")

                                @bind('#subm','click')
                                def welcome(event):
                                    event.preventDefault()
                                    req = ajax.Ajax()
                                    req.bind('complete', on_complete)
                                    accession = document['fname'].value
                                    req.open('POST','http://localhost:8000/', True)
                                    req.send({{'accession': accession, 'epicore_csv':'{epicore_csv}','proteome_dict':{ctx.obj.proteome_dict}}})
 
                            </script>

                            <form id="prot_landscape">
                                <label for="fname">Protein accession: </label>
                                <input type="text" id="fname" name="fname" list="exampleList"><br>
                                <datalist id="exampleList">
                                </datalist>
                                <button id="subm">Plot protein landscape</button>
                            </form> 
                            <div id="image_container"></div>
                            <p id="output"></p>
                            <p> The average length of an epitope core sequence is {mean_core_length}. <br> The average length of an epitope sequence is {mean_whole_length}.</p><br>
                            <p>{n_removed_peps} peptides were removed, since they do not appear in the provided evidence file.</p>
                            <p> The used parameter values:<br> - min_epi_length:{ctx.obj.min_epi_length}<br> 
                            - min_overlap:{ctx.obj.min_overlap}<br> - max_step_size:{ctx.obj.max_step_size}<br> - seq_column:{ctx.obj.seq_column}<br>
                            - protacc_column:{ctx.obj.protacc_column} <br> - intensity_column:{ctx.obj.intensity_column} <br> - start_column: {ctx.obj.start_column} <br>
                            - end_column: {ctx.obj.end_column} <br> - out_dir: {ctx.obj.out_dir} <br> - mod_pattern: {ctx.obj.mod_pattern} <br> - delimiter: {ctx.obj.delimiter} <br>
                            The input evidence file: {evidence_file}</p>
                            <p> The ten epitopes with the highest number of mapped peptides:<br>{epitope_df_sort_html}</p>
                            

                        </body>
                        </html>'''
    
    threading.Thread(target=epicore_utils.modules.locserver.run, args=(8000,html_content)).start()
    webbrowser.open_new_tab('http://localhost:8000/report.html')


import os
import pandas as pd 
import ast
import yaml
import click
import logging
import numpy as np 
import matplotlib.pyplot as plt
import warnings
import re
import polars as pl


from . import __version__
from epicore_utils.modules.compute_cores import compute_consensus_epitopes
from epicore_utils.modules.map_result import map_pep_core, gen_epitope_df
from epicore_utils.modules.visualize_protein import plot_protein_landscape, plot_peptide_length_dist, plot_core_mapping_peptides_hist, plot_consensus_sequence_coverage, create_html
from epicore_utils.modules.parse_input import parse_input, proteome_to_dict
from epicore_utils.modules.generate_report import gen_report

import logging

logger = logging.getLogger(__name__)

class InputParameter(object):
    """This class contains parameters necessary for the epicore script.

    Attributes:
        min_epi_length (int, optional): An integer of the minimal length of a 
            consensus epitope.
        min_overlap (int, optional): An integer of the minimal overlap between
            two epitopes to be grouped to the same consensus epitope.
        max_step_size (int, optional): An integer of the maximal distance 
            between the start position of two epitopes to be grouped to the same consensus epitope.
        seq_column (str, optional): The string of the header of the column 
            containing peptide sequence information in the evidence file.
        protacc_column (str, optional): The string of the header of the column  
            containing protein accession information in the evidence file.
        intensity_column (str, optional): The string of the header of the 
            column containing intensity information in the evidence file.
        delimiter (str, optional): The delimiter that separates multiple 
            entries in one column in the evidence file.
        mod_patter (str, optional): A comma separated string with delimiters
            for peptide modifications
        out_dir (str): A string of the directory, were all output files will be 
            saved.
        prot_accession (str, optional): A comma separated string containing the 
            protein accession, for which the protein landscape will be visualized.
        start_column (str, optional): The string of the header of the column 
            containing the start positions of peptides in proteins.
        end_column (str, optional): The string of the header of the column 
            containing the end position of peptides in proteins.

    """
    def __init__(self,reference_proteome=None, min_epi_length=None, min_overlap=None, max_step_size=None, seq_column=None, protacc_column=None, intensity_column=None, delimiter=None, mod_pattern=None, out_dir=None, prot_accession=None, start_column=None, end_column=None, report=None, html=None, sample_column=None, strict=None, condition=None):
        self.min_epi_length = min_epi_length
        self.min_overlap = min_overlap
        self.max_step_size = max_step_size
        self.seq_column = seq_column
        self.protacc_column = protacc_column
        self.intensity_column = intensity_column
        self.delimiter = delimiter
        self.mod_pattern = mod_pattern
        self.out_dir = out_dir
        self.prot_accession = prot_accession
        self.start_column = start_column
        self.end_column = end_column
        self.report = report
        self.html = html
        self.proteome_dict = proteome_to_dict(reference_proteome)
        self.reference_proteome = reference_proteome
        self.sample_column = sample_column
        self.strict = strict
        self.condition_column = condition

@click.version_option(__version__, "--version", "-V")

@click.group()
@click.option('--reference_proteome',type=click.Path(exists=True), required=True)
@click.option('--out_dir', type=click.Path(), required=True)
@click.pass_context
def main(ctx, reference_proteome, out_dir):
    ctx.obj = InputParameter(reference_proteome=reference_proteome, out_dir=out_dir)
    
@click.option('--min_epi_length', type=click.INT, default=11)
@click.option('--min_overlap', type=click.INT, default=11)
@click.option('--max_step_size', type=click.INT, default=5)
@click.option('--seq_column', type=click.STRING, required=True)
@click.option('--sample_column', type=click.STRING, required=True)
@click.option('--protacc_column', type=click.STRING, required=True)
@click.option('--condition_column', type=click.STRING, required=True)
@click.option('--intensity_column', type=click.STRING)
@click.option('--delimiter', type=click.STRING, required=True)
@click.option('--mod_pattern', type=click.STRING)
@click.option('--prot_accession', type=click.STRING)
@click.option('--start_column', type=click.STRING)
@click.option('--end_column', type=click.STRING)
@click.option('--report', is_flag=True)
@click.option('--html', is_flag=True)
@click.option('--strict', is_flag=True)
@click.command()
@click.option('--evidence_file',type=click.Path(exists=True), required=True)
@click.pass_context
def generate_epicore_csv(ctx,evidence_file, min_epi_length, min_overlap, max_step_size, seq_column, protacc_column, intensity_column, delimiter, mod_pattern, prot_accession, start_column, end_column, report, html, sample_column, strict, condition_column):
    ctx.obj = InputParameter(ctx.obj.reference_proteome, min_epi_length, min_overlap, max_step_size, seq_column, protacc_column, intensity_column, delimiter, mod_pattern, ctx.obj.out_dir, prot_accession, start_column, end_column, report, html, sample_column, strict, condition_column)
    if not os.path.exists(ctx.obj.out_dir):
        os.mkdir(ctx.obj.out_dir)
    logging.basicConfig(filename=f'{ctx.obj.out_dir}/epicore.log', level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
    logger.info(f'Parameter: min_epi_length:{min_epi_length}, min_overlap:{min_overlap}, max_step_size:{max_step_size}, strict:{strict}, evidence_file:{evidence_file}, fasta_file:{ctx.obj.reference_proteome}')
    
    # ----------------------
    #    Parse input file
    # ----------------------
    protein_df, n_removed_peps, total_intens = parse_input(evidence_file, ctx.obj.seq_column, ctx.obj.protacc_column, ctx.obj.intensity_column, ctx.obj.start_column, ctx.obj.end_column, ctx.obj.delimiter, ctx.obj.proteome_dict, ctx.obj.mod_pattern, ctx.obj.sample_column, ctx.obj.condition_column)
    protein_df = protein_df.to_pandas()
    os.makedirs(ctx.obj.out_dir,exist_ok=True)

    # ----------------------
    # compute core epitopes, map peptides to cores
    # ----------------------
    protein_df = compute_consensus_epitopes(protein_df, ctx.obj.min_overlap, ctx.obj.max_step_size, ctx.obj.min_epi_length, ctx.obj.intensity_column, ctx.obj.mod_pattern, ctx.obj.proteome_dict, total_intens, ctx.obj.strict)
    protein_df.to_csv(f'{ctx.obj.out_dir}/epicore_result.csv')
    pep_cores_mapping = map_pep_core(evidence_file,protein_df,ctx.obj.seq_column,ctx.obj.protacc_column,ctx.obj.start_column,ctx.obj.end_column,ctx.obj.intensity_column,ctx.obj.delimiter,ctx.obj.mod_pattern, ctx.obj.proteome_dict)
    pep_cores_mapping.to_csv(f'{ctx.obj.out_dir}/pep_cores_mapping.tsv', sep='\t', index=False)


    # ----------------------
    # Reformat data and generate multiple plots
    # ----------------------
    # generate file with one epitope in each row
    epitope_df = gen_epitope_df(protein_df)
    epitope_df.to_csv(f'{ctx.obj.out_dir}/epitopes.csv')

    plot_consensus_sequence_coverage(epitope_df, f'{ctx.obj.out_dir}')
    # compute length distribution of peptides and epitopes
    ext = os.path.splitext(evidence_file)[1]
    if ext == '.csv':
        evidence_df = pd.read_csv(evidence_file, delimiter=',')
    elif ext == '.tsv':
        evidence_df = pd.read_csv(evidence_file, delimiter='\t')
    elif ext == '.xlsx':
        evidence_df = pd.read_excel(evidence_file)
    evidence_df[ctx.obj.protacc_column] = evidence_df[ctx.obj.protacc_column].apply(lambda accessions: accessions.split(ctx.obj.delimiter))

    fig = plot_core_mapping_peptides_hist(epitope_df)
    fig.savefig(f'{ctx.obj.out_dir}/epitope_intensity_hist.svg')
    if ctx.obj.html:
        create_html(f'{ctx.obj.out_dir}/epitope_intensity_hist.html')

    fig, peps, epitopes = plot_peptide_length_dist(evidence_df, epitope_df, ctx.obj.seq_column, 'consensus_epitopes', ctx.obj.seq_column, 'consensus_epitopes', 'peptides', 'consensus epitopes', mod_pattern)
    fig.savefig(f'{ctx.obj.out_dir}/length_distributions.svg')
    if ctx.obj.html:
        create_html(f'{ctx.obj.out_dir}/length_distributions.html')
    
    # summarize some results
    if ctx.obj.report:
        gen_report(f'http://localhost:8000/{ctx.obj.out_dir}/length_distributions.svg', f'http://localhost:8000/{ctx.obj.out_dir}/epitope_intensity_hist.svg', epitope_df, peps, epitopes, n_removed_peps, ctx, evidence_file,  f'{ctx.obj.out_dir}/epicore_result.csv')


@click.command()
@click.option('--epicore_csv',type=click.Path(exists=True), required=True)
@click.option('--protacc', type=click.STRING, required=True)
@click.pass_context
def plot_landscape(ctx,epicore_csv, protacc):
    if not protacc:
        raise Exception('No protein accession was provided. Please provide a protein accession')
    for accession in protacc.split(','):

        # read in precomputed protein coverage and epitope cores.
        protein_df = pd.read_csv(epicore_csv)

        protein_df['grouped_peptides_start'] = protein_df['grouped_peptides_start'].apply(ast.literal_eval)
        protein_df['core_epitopes_start'] = protein_df['core_epitopes_start'].apply(ast.literal_eval)
        protein_df['core_epitopes_end'] = protein_df['core_epitopes_end'].apply(ast.literal_eval)
        protein_df['landscape'] = protein_df['landscape'].apply(ast.literal_eval)

        if accession is not None:
            fig = plot_protein_landscape(protein_df,accession,ctx.obj.proteome_dict)
            fig.savefig(f'{ctx.obj.out_dir}/{accession}.pdf',bbox_inches='tight')
            fig.savefig(f'{ctx.obj.out_dir}/{accession}.svg',bbox_inches='tight')
            

main.add_command(generate_epicore_csv)
main.add_command(plot_landscape)

if __name__ == '__main__':
    main()



"""
Assigns each peptide in the evidence files its core epitopes, the total intensity of that core epitope and the relative core intensity. 
"""
import warnings
import pandas as pd
import ast
import numpy as np
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)
import re
import os 
from itertools import repeat
import logging
import time
import polars as pl
logger = logging.getLogger(__name__)
from epicore_utils.modules.parse_input import read_entire_id_output

def aggregate_series(series: pd.Series, delimiter: str) -> str:
    """Aggregate the series.

    Args:
        series: A pandas series.
    
    Returns:
        A aggregated version of the pandas series. If all elements of the
        series are the same, only the first element is returned. If the 
        elements differ from each other a list of all elements is returned
        as a string.
    """
    series = series.fillna('nan')
    if series.name in  ['start', 'end', 'accession', 'whole_epitopes_all', 'proteome_occurrence']:
        series = series.astype(str)
        return delimiter.join(series)
    elif len(set(series)) == 1:
        return series.iloc[0]
    else:
        series = series.astype(str)
        return delimiter.join(series)


def map_pep_core(evidence_file: str, protein_df: pd.DataFrame, seq_column: str, protacc_column: str, start_column: str, end_column: str, intensity_column: str, delimiter: str, mod_pattern: str, proteome_dict: dict[str,str]) -> pd.DataFrame:
    """Map computed consensus epitope groups to the input evidence_file.
    
    Args:
        evidence_file: The string of the path to the evidence file.
        protein_df: A pandas dataframe containing one protein per row.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        delimiter: The delimiter that separates multiple entries in one column 
            in the evidence file.
        mod_pattern: A comma separated string with delimiters for peptide
            modifications

    Returns:
        The evidence_file with four additional columns containing the whole and 
        core sequence and total and relative intensity of each consensus 
        epitope group, to which the peptide of the row belongs.

    Raises:
        Exception: If the mappings are contradictory.
    """

    # read in entire evidence file
    evidence_file_df = read_entire_id_output(evidence_file, False)
    in_cols = evidence_file_df.columns.values

    if intensity_column:
        protein_df = protein_df[['accession', 'start', 'end', 'intensity', 'whole_epitopes_all','consensus_epitopes_all', 'core_epitopes_intensity_all', 'relative_core_intensity_all', 'proteome_occurrence', 'peptide_index']]
    else:
        protein_df = protein_df[['accession', 'start', 'end', 'whole_epitopes_all','consensus_epitopes_all', 'proteome_occurrence', 'peptide_index']]

    # reformat protein_df so every peptide sequence is represented by one row
    if intensity_column:
        protein_df = protein_df.explode(['start', 'end', 'intensity', 'whole_epitopes_all','consensus_epitopes_all', 'core_epitopes_intensity_all', 'relative_core_intensity_all', 'proteome_occurrence', 'peptide_index'])
    else: 
        protein_df = protein_df.explode(['start', 'end', 'whole_epitopes_all','consensus_epitopes_all', 'proteome_occurrence', 'peptide_index'])
    
    # prepare protein_df
    protein_df['peptide_index'] = protein_df['peptide_index'].str.replace('[','').str.replace(']','').str.replace(' ','').str.split(',')
    protein_df = protein_df.explode(['peptide_index'])
    protein_df['start'] = protein_df['start'].astype(int)
    protein_df = protein_df.sort_values(by=['accession', 'start'])
    protein_df['start'] = protein_df['start'].astype(str)
    protein_df = protein_df.groupby('peptide_index').agg(lambda x: delimiter.join(x))

    # prepare peptide_df
    if start_column and end_column:
        evidence_file_df = evidence_file_df.drop(columns=['start', 'end', protacc_column])
    else: 
        evidence_file_df = evidence_file_df.drop(columns=[protacc_column])
    evidence_file_df['index'] = evidence_file_df.reset_index()['index'].astype(str)

    # merge dataframes on peptide index, use the input columns
    evidence_file_df = evidence_file_df.reset_index()
    protein_df = protein_df.reset_index()
    evidence_file_df = pd.merge(evidence_file_df, protein_df, left_on=['index'], right_on=['peptide_index'], how='right')
    rename={'whole_epitopes_all':'entire_epitope_sequence','consensus_epitopes_all':'consensus_epitope_sequence', 'sequence':seq_column, 'accession':protacc_column}
    if intensity_column:
        rename['intensity'] = intensity_column
    evidence_file_df = evidence_file_df.rename(columns=rename)
    drop_cols = ['index', 'peptide_index']
    evidence_file_df = evidence_file_df.drop(drop_cols, axis=1)
    evidence_file_df.columns = evidence_file_df.columns.str.replace(r'_y$','', regex=True)

    # sort the dataframe columns so they match the input order
    out_cols = list(in_cols) + [col for col in evidence_file_df.columns.values if col not in in_cols]
    evidence_file_df = evidence_file_df[out_cols].rename(columns=rename)
    if intensity_column:
        evidence_file_df = evidence_file_df.rename(columns={'core_epitopes_intensity_all':'consensus_epitope_intensity','relative_core_intensity_all':'relative_consensus_epitope_intensity' })

    evidence_file_df = evidence_file_df.drop(columns='level_0')
    return evidence_file_df

def gen_epitope_df(protein_df: pd.DataFrame) -> pd.DataFrame:
    """Generate dataframe that has one epitope per row.

    Args:
        protein_df: A pandas dataframe containing one protein per row.

    Returns:
        A reordered version of protein_df were each row stores one epitope.
    """
    # include intensity columns if present
    if ('core_epitopes_intensity' not in protein_df.columns) and ('relative_core_intensity' not in protein_df.columns):
        cols = ['whole_epitopes', 'consensus_epitopes','landscape', 'grouped_peptides_sequence', 'grouped_peptides_sample', 'grouped_peptides_condition', 'grouped_peptides_start', 'grouped_peptides_end','core_epitopes_start', 'core_epitopes_end']
    else:
        cols = ['whole_epitopes', 'consensus_epitopes','landscape', 'grouped_peptides_sequence', 'relative_core_intensity', 'core_epitopes_intensity']

    cols_acc = cols + ['accession']

    # separate each epitope in one row
    protein_df_long = protein_df[cols_acc].explode(cols)
    protein_df_long = protein_df_long.astype(str)

    # remove peptide groups occurring multiple times
    protein_df_long = protein_df_long.drop_duplicates()

    epitopes_grouped_df = protein_df_long.groupby(cols)
    epitopes_grouped_df = epitopes_grouped_df.agg({'accession':lambda x:','.join(x)}).reset_index()

    logger.info(f'{len(epitopes_grouped_df.drop_duplicates(["whole_epitopes","consensus_epitopes"]))} unique epitopes were computed.')

    return epitopes_grouped_df
